<?php

namespace Drupal\monarch_entity_yml\Commands;

use Drush\Commands\DrushCommands;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Config\Entity\ConfigEntityInterface;

/**
 * A Drush commandfile.
 *
 * In addition to this file, you need a drush.services.yml
 * in root of your module, and a composer.json file that provides the name
 * of the services file to use.
 *
 * See these files for an example of injecting Drupal services:
 *   - http://cgit.drupalcode.org/devel/tree/src/Commands/DevelCommands.php
 *   - http://cgit.drupalcode.org/devel/tree/drush.services.yml
 */
class EntityYmlCommands extends DrushCommands {

  /**
   * All references.
   *
   * @var array
   */
  protected $references = [];

  /**
   * All entities exported.
   *
   * @var array
   */
  protected $entities = [];

  /**
   * Export data.
   *
   * @var array
   */
  protected $data = [];

  /**
   * Export order.
   *
   * @var array
   */
  protected $order = [];

  /**
   * Referenced allowed to be exported.
   *
   * @var array
   */
  protected $exportReferenceTypes = [];

  /**
   * Include non-banned references.
   *
   * @var bool
   */
  protected $includeReferences = FALSE;

  /**
   * A reference of fields that shouldn't be exported.
   *
   * @var array
   */
  protected const BANNED_FIELDS = [
    'any' => [
      'path' => TRUE,
    ],
  ];

  /**
   * Add ancestor entities.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to attempt loading ancestors for.
   * @param bool $export_if_neutral
   *   Export the entity by default unless specifically blacklisted.
   * @param int $level
   *   The reference level/weight.
   */
  protected function queueEntity(ContentEntityInterface $entity, bool $export_if_neutral = FALSE, int $level = 0) {
    if (isset($this->entities[$entity->getEntityTypeId() . ':' . $entity->id()])) {
      return;
    }

    $this->entities[$entity->getEntityTypeId() . ':' . $entity->id()] = TRUE;
    $this->order[$entity->getEntityTypeId()] = min($this->order[$entity->getEntityTypeId()] ?? 9999, $level * 10);

    if (method_exists($entity, 'getParentId')) {
      $parentId = $entity->getParentId();

      if (!empty($parentId)) {
        list($entityTypeId, $uuid) = explode(':', $parentId);

        $parentEntity = \Drupal::service('entity.repository')->loadEntityByUuid($entityTypeId, $uuid);

        if (!empty($parentEntity)) {
          $this->queueEntity($parentEntity, $this->includeReferences, $level - 1);
        }
      }
    }

    if (!method_exists($entity, 'getFields')) {
      return;
    }

    $entity_data = [
      'storage' => $entity->getEntityTypeId(),
      'weight' => $level,
    ];

    foreach ($entity->getFields() as $name => $field) {
      if (
        (static::BANNED_FIELDS['any'][$name] ?? FALSE) ||
        (static::BANNED_FIELDS[$entity->getEntityTypeId()][$name] ?? FALSE)
      ) {
        continue;
      }

      $entity_data['values'][$name] = $field->getValue();

      if (method_exists($field, 'referencedEntities')) {
        foreach ($field->referencedEntities() as $childEntity) {
          $childKey = $childEntity->getEntityTypeId() . ':' . $childEntity->id();

          if (!($childEntity instanceof ConfigEntityInterface)) {
            $this->references[$childKey] = $childKey;
          }

          if ($childEntity instanceof ContentEntityInterface) {
            if ($export_if_neutral && !isset($this->exportReferenceTypes[$childEntity->getEntityTypeId()]) || $this->exportReferenceTypes[$childEntity->getEntityTypeId()] ?? NULL) {
              $this->queueEntity($childEntity, $this->includeReferences, $level - 1);
            }
          }
        }
      }
    }

    if ($entity instanceof ContentEntityInterface) {
      if ($export_if_neutral && !isset($this->exportReferenceTypes[$entity->getEntityTypeId()]) || $this->exportReferenceTypes[$entity->getEntityTypeId()] ?? NULL) {
        array_push($this->data, $entity_data);
      }
    }
  }

  /**
   * Export entity as configuration.
   *
   * @param string $entity_list
   *   Entity ids.
   * @param array $options
   *   The options provided by drush.
   *
   * @command entity_yml:dump
   *
   * @aliases eyd
   * @option $with-references Include all references (unless explicitly excluded).
   * @option $file Output to file instead of stdout. If filename is not specified then it will be generated automatically.
   * @option $path Override output directory (only when writing to file). If not specified then the site's content directory will be looked for.
   * @option $prefix Prefix to be prepended to the filename.
   * @usage entity_yml:dump entity_type:id > file.yml
   *   Export an entity to yml.
   * @usage entity_yml:dump entity_type:id,id,id... > file.yml
   *   Export multiple entities of a type to yml.
   * @usage entity_yml:dump entity_type:* > file.yml
   *   Export all entities of a type to yml.
   * @usage entity_yml:dump entity_type:+ > file.yml
   *   Export all secondary references to entities of a type to yml.
   * @usage entity_yml:dump entity_type:- > file.yml
   *   Exclude entity type from export (but still check it for references, if
   *   included by other criteria).
   * @usage entity_yml:dump ...;...;... > file.yml
   *   Specify multiple criteria.
   * @usage entity_yml:dump "... ... ..." > file.yml
   *   Specify multiple criteria.
   * @usage entity_yml:dump "entity_type:id another_entity_type:* another_entity_type:+" > file.yml
   *   Export multiple entities and their referenced entities to yml.
   */
  public function entityExport(string $entity_list = NULL, array $options = [
    'with-references' => FALSE,
    'path' => FALSE,
    'file' => '',
    'prefix' => '000',
  ]) {
    $entityTypeManager = \Drupal::service('entity_type.manager');

    $this->includeReferences = $options['with-references'] ? TRUE : FALSE;
    $this->data = [];
    $this->references = [];
    $this->entities = [];
    $this->exportReferenceTypes = [];

    preg_match_all('/([^\s;:]+):([^\s;:]+)/i', $entity_list, $matches, PREG_SET_ORDER);

    foreach ($matches as $match) {
      if ($match[2] === '*' || $match[2] === '+') {
        $this->exportReferenceTypes[$match[1]] = $this->exportReferenceTypes[$match[1]] ?? TRUE;
      }
      elseif ($match[2] === '-') {
        $this->exportReferenceTypes[$match[1]] = FALSE;
      }
    }

    foreach ($matches as $match) {
      try {
        $storage = $entityTypeManager->getStorage($match[1]);

        if (empty($storage)) {
          $this->stderr()->writeln('Entity storage not found: ' . $match[1]);
          continue;
        }

        if ($match[2] === '*') {
          $res = $storage->loadMultiple();
        }
        elseif ($match[2] === '+') {
          continue;
        }
        else {
          $res = $storage->loadMultiple(explode(',', $match[2]));
        }

        foreach ($res as $entity) {
          if (!($entity instanceof ContentEntityInterface)) {
            $this->stderr()->writeln('Skipping non-content entity: ' . $entity->getEntityTypeId() . ':' . $entity->id());
            continue;
          }

          if (empty($entity)) {
            $this->stderr()->writeln('Entity not found: ' . $match[0]);
            continue;
          }

          $this->queueEntity($entity, TRUE);
        }
      }
      catch (\Throwable $e) {
        $this->stderr()->writeln($e->getMessage());
      }
    }

    usort($this->data, function ($a, $b) {
      return ($a['weight'] ?? 9999) - ($b['weight'] ?? 9999);
    });

    if ($options['file']) {
      $fs = \Drupal::service('file_system');
      if (!empty($options['path']) && is_string($options['path'])) {
        $content_path = $fs->realpath($options['path']);

        if (!$content_path) {
          $this->stderr()->writeln('Path not found. Please create the directory and ensure that it is writable.');
          return;
        }
      }
      else {
        $content_path = $fs->realpath(drupal_get_path('profile', \Drupal::installProfile()) . '/content')
          ?: $fs->realpath('../content')
          ?: $fs->realpath('content');

        if (!$content_path) {
          $this->stderr()->writeln('Content directory not found. Please create a \'content\' directory either in your install profile or at the root of your installation.');
          return;
        }
      }

      $filename = is_string($options['file']) ? $options['file'] : implode('--', array_unique(array_map(function ($data) {
        return $data['storage'];
      }, $this->data)));

      if (!empty($options['prefix']) && is_string($options['prefix'])) {
        $filename = $options['prefix'] . '--' . $filename;
      }

      $filename = preg_replace('/[^a-z0-9\-]/i', '-', $filename) . '.yml';

      foreach ($this->data as &$entry) {
        $uuid = $entry['values']['uuid'][0]['value'] ?? NULL;

        if ($uuid) {
          $uuid = '_' . $uuid;
          $entryfile = $uuid . '.yml';
          file_put_contents($content_path . '/' . $entryfile, Yaml::encode($entry));
          $entry = $uuid;
        }
      }

      file_put_contents($content_path . '/' . $filename, Yaml::encode($this->data));
    }
    else {
      $this->output()->writeln(Yaml::encode($this->data));
    }

    if (!empty($this->references)) {
      $this->references = array_filter(array_values($this->references), function ($key) {
        return empty($this->entities[$key]);
      });

      $this->stderr()->writeln("Unexported References: " . implode(' ', $this->references));
    }
  }

  /**
   * Export entity as configuration.
   *
   * @command entity_yml:import
   *
   * @aliases eyi
   * @option $path Load yml files from this directory instead of stdin. If not specified then the site's content directory will be looked for.
   * @usage entity_yml:import < file.yml
   *  Imports yml and attempts to recreate entities defined within.
   */
  public function entityImport(array $options = [
    'path' => FALSE,
  ]) {
    $fs = \Drupal::service('file_system');

    if ($options['path'] !== FALSE) {
      if (is_string($options['path'])) {
        $content_path = $fs->realpath($options['path']);

        if (!$content_path) {
          throw new \Exception('Path not found. Please create the directory and ensure that it is writable.');
        }
      }
      else {
        $content_path = $fs->realpath(drupal_get_path('profile', \Drupal::installProfile()) . '/content')
          ?: $fs->realpath('../content')
          ?: $fs->realpath('content');

        if (!$content_path) {
          throw new \Exception('Content directory not found. Please create a \'content\' directory either in your install profile or at the root of your installation.');
        }
      }

      try {
        $feedback = \Drupal::service('monarch_entity_yml.import')->import($content_path);

        $this->output()->writeln('Entities created:');

        foreach ($feedback as $entry) {
          $this->output()->writeln($entry);
        }
      }
      catch (\Throwable $e) {
        $this->stderr()->writeln($e->getMessage());
        return;
      }
    }
    else {
      try {
        $data = Yaml::decode(stream_get_contents(STDIN));
        $feedback = \Drupal::service('monarch_entity_yml.import')->importData($data);

        $this->output()->writeln('Entities created:');

        foreach ($feedback as $entry) {
          $this->output()->writeln($entry);
        }
      }
      catch (\Throwable $e) {
        $this->stderr()->writeln($e->getMessage());
        return;
      }
    }
  }

  /**
   * List entity types eligible for export.
   *
   * @usage entity_yml:list
   *  List entities eligible for export.
   *
   * @command entity_yml:list
   * @aliases eyl
   */
  public function entityList() {
    $entityTypeManager = \Drupal::service('entity_type.manager');
    $entity_definitions = $entityTypeManager->getDefinitions();

    $list = [];

    foreach ($entity_definitions as $definition) {
      if ($definition instanceof ContentEntityTypeInterface) {
        $list[] = $definition->id();
      }
    }

    if (empty($list)) {
      $this->stderr()->writeln('No entity types eligible for export.');
    }
    else {
      $this->output()->writeln('Entity types eligible for export:');

      sort($list);

      foreach ($list as $entity_type) {
        $this->output()->writeln("\t" . $entity_type);
      }
    }
  }

}
